import re
from keyword import iskeyword
from typing import List


def is_valid_key(key: str) -> bool:
    if iskeyword(key):
        return False
    return key.isidentifier()


def key_to_class_name(key: str) -> str:
    # First split on non characters
    parts1 = re.split(r"[^a-zA-Z0-9]", key)

    # Then split each if camelcase
    parts2: List[str] = []
    for part in parts1:
        if part.islower():
            parts2.append(part)
            continue

        # Assume it's pascal or camel case
        for sub_part in re.split(r"([A-Z][^A-Z]+)", part):
            if len(sub_part):
                parts2.append(sub_part)

    return "".join([part[0].upper() + part[1:].lower() for part in parts2 if part])



import functools
from typing import Any, Dict, List, Optional, Set, TypeVar

from dict_typer.utils import is_valid_key

KNOWN_TYPE_IMPORTS = ("List", "Tuple", "Set", "FrozenSet")


EntryType = TypeVar("EntryType", "MemberEntry", "DictEntry")
SubMembers = Set[EntryType]
DictMembers = Dict[str, SubMembers]


def is_valid_name(name: str) -> bool:
    if not is_valid_key(name):
        return False
    return name not in KNOWN_TYPE_IMPORTS


def sub_members_to_string(sub_members: SubMembers) -> str:
    def get_member_value(item: EntryType) -> str:
        """ Only reference DictEntry by name. """
        if isinstance(item, DictEntry):
            return item.name
        return str(item)

    if len(sub_members) == 2 and "None" in (sm.name for sm in sub_members):
        optional_member = next(sm for sm in sub_members if sm.name != "None")
        return f"Optional[{get_member_value(optional_member)}]"
    if len(sub_members) > 0:
        sub_members_strs = sorted(get_member_value(sm) for sm in sub_members)
        if len(sub_members_strs) == 1:
            return str(sub_members_strs[0])
        return f"Union[{', '.join(sub_members_strs)}]"
    return ""


def sub_members_to_imports(sub_members: SubMembers) -> Set[str]:
    imports = set()

    for member in sub_members:
        imports |= member.get_imports()

    if len(sub_members) == 2 and "None" in (sm.name for sm in sub_members):
        imports.add("Optional")
    elif len(sub_members) > 1:
        imports.add("Union")

    return imports


class MemberEntry:
    """ A representation of a type with optional sub types.

    A MemberEntry without subtypes can be considered as a leaf of a tree, in
    most cases it will be a simple unit such as str, int, float, but it can
    also be a List, Set, Tuple, Dict without any known members.
    """

    name: str
    sub_members: SubMembers

    def __init__(self, name: str, sub_members: Optional[SubMembers] = None,) -> None:
        self.name = name
        self.sub_members = sub_members or set()

    def get_imports(self) -> Set[str]:
        imports = set()
        if self.name in KNOWN_TYPE_IMPORTS:
            imports.add(self.name)

        return imports | sub_members_to_imports(self.sub_members)

    @property
    def depends_on(self) -> Set[str]:
        return {sm.name for sm in self.sub_members}

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        if self.__class__ != other.__class__:
            return False
        assert isinstance(other, self.__class__)

        return str(self) == str(other)

    def __repr__(self) -> str:
        return f"<MemberEntry ({self})>"

    def __str__(self) -> str:
        sub_string = sub_members_to_string(self.sub_members)
        if sub_string:
            return f"{self.name}[{sub_string}]"

        return self.name


class DictEntry:
    """ A representation of a typed dict.

    A typed dict will have a name and a members map. The value of each member
    is a MemberEntry, which has a name and an optional submembers.
    """

    name: str
    members: DictMembers
    indentation: int
    force_alternative: bool

    def __init__(
        self,
        name: str,
        members: Optional[DictMembers] = None,
        indentation: int = 4,
        force_alternative: bool = False,
    ) -> None:
        if is_valid_name(name):
            self.name = name
        else:
            self.name = f"{name}_"
        self.members = members or {}
        self.indentation = indentation
        self.force_alternative = force_alternative

    def get_imports(self) -> Set[str]:
        imports = set()
        for sub_members in self.members.values():
            imports |= sub_members_to_imports(sub_members)
        return imports

    def update_members(self, members: DictMembers) -> None:
        if set(members.keys()) != self.keys:
            raise Exception("Keys don't match between members")

        for key, value in self.members.items():
            value |= members[key]

    def any_invalid_key(self) -> bool:
        return any(not is_valid_key(key) for key in self.keys)

    @property
    def keys(self) -> Set[str]:
        return set(self.members.keys())

    @property
    def depends_on(self) -> Set[str]:
        if not self.members:
            return set()
        members = set.union(*self.members.values())
        return set.union(*[m.depends_on for m in members], {m.name for m in members})

    def __hash__(self) -> int:
        return hash(str(";".join(self.keys)))

    def __eq__(self, other: Any) -> bool:
        """ DictEntries are equal if the keys are equal.

        The name and the values of each key don't matter, since the the name is
        just for the python type and the values are to know what type the keys
        can be, but the keys themselves are the only important thing
        """
        if self.__class__ != other.__class__:
            return False
        assert isinstance(other, self.__class__)

        return hash(self) == hash(other)

    def __repr__(self) -> str:
        return f"<DictEntry ({self.name})>"

    def __str__(self) -> str:
        out: List[str] = []

        if self.force_alternative or self.any_invalid_key():
            if not self.members:
                out.append(f'{self.name} = TypedDict("{self.name}", {{}}')
            else:
                out.append(f'{self.name} = TypedDict("{self.name}", {{')

                for key, value in self.members.items():
                    out.append(
                        f'{" " * self.indentation}"{key}": {sub_members_to_string(value)},'
                    )

                out.append("})")
        else:
            out.append(f"class {self.name}(TypedDict):")
            if not self.members:
                out.append(f'{" " * self.indentation}pass')
            else:
                for key, value in self.members.items():
                    out.append(
                        f"{' ' * self.indentation}{key}: {sub_members_to_string(value)}"
                    )

        return "\n".join(out)


@functools.total_ordering
class DependencyCmp:
    _name: str
    _depends_on: Set[str]

    def __init__(self, obj: EntryType) -> None:
        self._name = obj.name
        self._depends_on = obj.depends_on

    def __lt__(self, other: Any) -> bool:
        """ Considered less if other depends on it """
        if self.__class__ != other.__class__:
            raise TypeError(
                "DependencyCmp comparison only supports comparing to other instances of DependencyCmp"
            )
        assert isinstance(other, self.__class__)

        return self._name in other._depends_on

    def __eq__(self, other: Any) -> bool:
        """ Considered equal if other doesn't depend on it """
        if self.__class__ != other.__class__:
            raise TypeError(
                "DependencyCmp comparison only supports comparing to other instances of DependencyCmp"
            )
        assert isinstance(other, self.__class__)

        return self._name not in other._depends_on


def key_to_dependency_cmp(entry: Any) -> DependencyCmp:
    return DependencyCmp(entry)



from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union

from dict_typer.models import (
    DictEntry,
    MemberEntry,
    key_to_dependency_cmp,
    sub_members_to_imports,
    sub_members_to_string,
)
from dict_typer.utils import key_to_class_name

BASE_TYPES: Tuple[Type, ...] = (  # type: ignore
    bool,
    bytearray,
    bytes,
    complex,
    enumerate,
    float,
    int,
    memoryview,
    range,
    str,
    type,
    filter,
    map,
    zip,
)


Source = Union[str, int, float, bool, None, Dict, List]
NameMap = Dict[str, str]


class DefinitionBuilder:
    definitions: List[DictEntry]
    root_type_name: str
    type_postfix: str
    show_imports: bool
    source: Source
    name_map: NameMap

    _output: Optional[str] = None

    def __init__(
        self,
        source: Source,
        *,
        root_type_name: str = "Root",
        type_postfix: str = "",
        show_imports: bool = True,
        force_alternative: bool = False,
        name_map: NameMap = None,
    ) -> None:
        self.definitions = []

        self.root_type_name = root_type_name
        self.type_postfix = type_postfix
        self.show_imports = show_imports
        self.force_alternative = force_alternative
        if name_map:
            self.name_map = name_map
        else:
            self.name_map = {}

        self.source = source

    def _get_name(self, name: str) -> str:
        """ Return the mapped name if it exist """
        return self.name_map.get(name, name)

    def _add_definition(self, entry: DictEntry) -> DictEntry:
        """ Add an entry to the definions.

        If the entry is a DictEntry and there's an existing entry with the same
        keys, then combine the DictEntries
        """
        dicts_only = [d for d in self.definitions if isinstance(d, DictEntry)]
        for definition in dicts_only:
            if entry.keys == definition.keys:
                definition.update_members(entry.members)
                return definition
            if entry.name == definition.name:
                idx = 1
                new_name = f"{entry.name}{idx}"
                dicts_names = [d.name for d in dicts_only]
                while new_name in dicts_names:
                    idx += 1
                    new_name = f"{entry.name}{idx}"
                entry.name = new_name
        self.definitions.append(entry)
        return entry

    def _convert_list(self, key: str, lst: List, item_name: str) -> MemberEntry:
        entry = MemberEntry(key)

        idx = 0
        for item in lst:
            item_type = self._get_type(item, key=f"{item_name}{idx}")

            entry.sub_members.add(item_type)
            if isinstance(item_type, DictEntry):
                self._add_definition(item_type)
            idx += 1

        return entry

    def _convert_dict(self, type_name: str, dct: Dict) -> DictEntry:
        entry = DictEntry(
            self._get_name(type_name), force_alternative=self.force_alternative
        )
        for key, value in dct.items():
            value_type = self._get_type(value, key=key)
            if isinstance(value_type, DictEntry):
                definition = self._add_definition(value_type)
                value_type = definition
            entry.members[key] = {value_type}
        return entry

    def _get_type(self, item: Any, key: str) -> Union[MemberEntry, DictEntry]:
        if item is None:
            return MemberEntry("None")

        if isinstance(item, BASE_TYPES):
            return MemberEntry(type(item).__name__)

        if isinstance(item, (list, set, tuple, frozenset)):
            if isinstance(item, list):
                sequence_type_name = "List"
            elif isinstance(item, set):
                sequence_type_name = "Set"
            elif isinstance(item, frozenset):
                sequence_type_name = "FrozenSet"
            else:
                sequence_type_name = "Tuple"

            list_item_types: Set[Union[MemberEntry, DictEntry]] = set()
            idx = 0
            for value in item:
                item_type = self._get_type(value, key=f"{key}Item{idx}")
                if isinstance(item_type, DictEntry):
                    if item_type.members not in (
                        getattr(t, "members", None) for t in list_item_types
                    ):
                        list_item_types.add(item_type)
                        idx += 1
                        if isinstance(item_type, DictEntry):
                            self._add_definition(item_type)
                else:
                    list_item_types.add(item_type)

            return MemberEntry(sequence_type_name, sub_members=list_item_types)

        if isinstance(item, dict):
            return self._convert_dict(
                f"{key_to_class_name(key)}{self.type_postfix}", item
            )

        raise NotImplementedError(f"Type handling for '{type(item)}' not implemented")

    def build_output(self) -> str:
        if self._output:
            return self._output

        source_type = self._get_type(self.source, key=self.root_type_name)
        if isinstance(source_type, DictEntry):
            self._add_definition(source_type)
            root_item = None
        else:
            root_item = source_type

        # Convert the definitions to structured output
        self._output = ""

        if self.show_imports:
            typing_imports = set()
            typed_dict_import = False

            for definition in self.definitions:
                if isinstance(definition, DictEntry):
                    typed_dict_import = True
                typing_imports |= definition.get_imports()
            if root_item:
                typing_imports |= sub_members_to_imports({root_item})

            if typing_imports:
                self._output += "\n".join(
                    [f"from typing import {', '.join(sorted(typing_imports))}", "", ""]
                )
            if typed_dict_import:
                self._output += "\n".join(
                    ["from typing_extensions import TypedDict", "", "", ""]
                )

        self._output += "\n\n\n".join(
            [str(d) for d in sorted(self.definitions, key=key_to_dependency_cmp)]
        )

        if not isinstance(source_type, DictEntry):
            if len(self._output):
                self._output += "\n"
                if len(self.definitions):
                    self._output += "\n\n"
            self._output += f"{self.root_type_name}{self.type_postfix} = {sub_members_to_string({source_type})}"

        return self._output


def get_type_definitions(
    source: Source,
    root_type_name: str = "Root",
    type_postfix: str = "",
    show_imports: bool = True,
    force_alternative: bool = False,
    name_map: NameMap = None,
) -> str:
    builder = DefinitionBuilder(
        source,
        root_type_name=root_type_name,
        type_postfix=type_postfix,
        show_imports=show_imports,
        force_alternative=force_alternative,
        name_map=name_map,
    )

    return builder.build_output()
